/* ui_plugin_cmd.c - implementation of a command-line plugin for r2l. */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

#include "ui_plugin.h"

/* plugin's data structure. */
struct ui_plugin_data {
    FILE* fout;			/* file to print r2l mode into.		  */
    char* progname;		/* name of the program we're pluged into. */
    char* title;		/* title used as output's prefix.         */
    int thin_interface;		/* should we show a tin interface or not. */
    R2L_STATE md;		/* last known r2l mode.                   */
    BIDITEXT_BASE_STATE bs;	/* last known biditext base mode.         */
};

/* print 'help' information on screen. */
static
void
ui_show_help(ui_plugin_t ui)
{
    assert(ui);

    fprintf(ui->fout, "\r\n");
    fprintf(ui->fout, "%s key bindings:\r\n", ui->progname);
    fprintf(ui->fout, "\r\n");
    fprintf(ui->fout, "Press 'r' to enable biditext, 'l' to disbale.\r\n");
    fprintf(ui->fout, "Press '1' to set bidi base state to LTR.\r\n");
    fprintf(ui->fout, "Press '2' to set bidi base state to Neutral.\r\n");
    fprintf(ui->fout, "Press '3' to set bidi base state to RTL.\r\n");
    fprintf(ui->fout, "Press 'q' to quit.\r\n");
    fprintf(ui->fout, "Press 'h' to get this help message.\r\n");
    fprintf(ui->fout, "\r\n");
    fflush(ui->fout);
}

/* initialize a given client. returns a ui_plugin_t opaque pointer. */
ui_plugin_t
ui_init(int argc,
	char* argv[],
	r2llib_t r2l_ctrl,
	char* title,
	int thin_interface)
{
    ui_plugin_t ui = (ui_plugin_t)malloc(sizeof(struct ui_plugin_data));

    if (!ui)
	goto ui_init_ret;

    ui->fout = stdout;
    ui->md = R2L_ERROR;
    ui->bs = BT_BASE_ERROR;
    ui->thin_interface = thin_interface;
    ui->progname = (char*)malloc(strlen(argv[0])+1);
    if (!ui->progname)
	goto err_free_ui_plugin;
    strcpy(ui->progname, argv[0]);
    ui->title = (char*)malloc(strlen(title)+1);
    if (!ui->title)
	goto err_free_progname;
    strcpy(ui->title, title);

    if (!ui->thin_interface) {
        fprintf(ui->fout,
		"%s: Terminal Mode Plugin Initialized\n", ui->progname);
	ui_show_help(ui);
        fflush(ui->fout);
    }

    system("stty raw -echo isig");
    system("stty onlret");

    goto ui_init_ret;

err_free_progname:
    free(ui->progname);
    ui->progname = NULL;
    
err_free_ui_plugin:
    free(ui);
    ui = NULL;
    
ui_init_ret:
    return ui;
}

/* define whether the ui plugin runs its own main loop. */
/* returns '1' for true, '0' for false.                 */
/* note that if it does, it must also support the timer */
/* registration interface.                              */
int
ui_provides_main_loop(ui_plugin_t ui)
{
    /* we don't provide our own main loop. */
    return 0;
}

/* execute the ui plugin's main loop - invoked only if */
/* 'ui_provides_main_loop' returned '1' previously.    */
void
ui_main_loop(ui_plugin_t ui)
{
    /* do nothing - we don't provide our own main loop. */
}

/* register a callback with the ui plugin to be invoked */
/* every 'ms' milliseconds.                             */
/* invoked only if 'ui_provides_main_loop' returned '1' */
/* previously.                                          */
int
ui_register_timer_callback(ui_plugin_t ui,
			   int ms,
			   int (*cb)(void*),
			   void* data)
{
    /* do nothing - we don't provide our own main loop, */
    /* and thus don't need to support timers.           */

    return 0;
}

/* notify the ui plugin about a mode change. */
void
ui_notify_of_r2l_mode(ui_plugin_t ui, R2L_STATE md, BIDITEXT_BASE_STATE bs)
{
    if (md != ui->md) {
	ui->md = md;
        fprintf(ui->fout,
	        "%s: %s mode: %s\r\n",
	        ui->progname,
	        (ui->thin_interface ? "" : ui->title),
	        (md == R2L_DISABLED ? "l" :
		    (md == R2L_ENABLED ? "r" : "unknown")));
    }
    if (bs != ui->bs) {
	ui->bs = bs;
        fprintf(ui->fout,
	        "%s: %s biditext mode: %s\r\n",
	        ui->progname,
	        (ui->thin_interface ? "" : ui->title),
	        (bs == BT_BASE_NEUTRAL ? "n" :
		    (bs == BT_BASE_LTR ? "ltr" :
			(bs == BT_BASE_RTL ? "r2l" : "unknown"))));
    }
}

/* returns a file descriptor that the UI would like r2l to listen on. */
/* for the command-line plugin - this is stdin.			      */
int
ui_get_listening_fd(ui_plugin_t ui)
{
    return fileno(stdin);
}

/* callback invoked when r2l discovers the file descriptor of the UI is active.
*/
int 
ui_perform_callback(ui_plugin_t ui, r2llib_t r2l_ctrl)
{
    int c = fgetc(stdin);
    int rc;

    if (tolower(c) == 'l') {
	rc = r2l_disable(r2l_ctrl);
	if (rc == 1) {
	    fprintf(ui->fout,
		    "%s: %s r2l mode set to 'l'\r\n",
		    ui->progname,
		    ui->thin_interface ? "" : ui->title);
	    fflush(ui->fout);
	    ui->md = R2L_DISABLED;
	}
    }
    else if (tolower(c) == 'r') {
	rc = r2l_enable(r2l_ctrl);
	if (rc == 1) {
	    fprintf(ui->fout,
		    "%s: %s r2l mode set to 'r'\r\n",
		    ui->progname,
		    ui->thin_interface ? "" : ui->title);
	    fflush(ui->fout);
	    ui->md = R2L_ENABLED;
	}
    }
    else if (c == '1') {
	rc = r2l_set_biditext_base_state_ltr(r2l_ctrl);
	if (rc == 1) {
	    fprintf(ui->fout,
		    "%s: %s biditext base mode set to 'ltr'\r\n",
		    ui->progname,
		    ui->thin_interface ? "" : ui->title);
	    fflush(ui->fout);
	    ui->bs = BT_BASE_LTR;
	}
    }
    else if (c == '2') {
	rc = r2l_set_biditext_base_state_neutral(r2l_ctrl);
	if (rc == 1) {
	    fprintf(ui->fout,
		    "%s: %s biditext base mode set to 'neutral'\r\n",
		    ui->progname,
		    ui->thin_interface ? "" : ui->title);
	    fflush(ui->fout);
	    ui->bs = BT_BASE_NEUTRAL;
	}
    }
    else if (c == '3') {
	rc = r2l_set_biditext_base_state_rtl(r2l_ctrl);
	if (rc == 1) {
	    fprintf(ui->fout,
		    "%s: %s biditext base mode set to 'rtl'\r\n",
		    ui->progname,
		    ui->thin_interface ? "" : ui->title);
	    fflush(ui->fout);
	    ui->bs = BT_BASE_RTL;
	}
    }
    else if (tolower(c) == 'h') {
	ui_show_help(ui);
    } 
    else if (tolower(c) == 'q') {
	return 1; /* we're done */
    }
    return 0; 
}

/* cleanup the ui plugin. after this call, 'ui' may no longer be used. */
void
ui_cleanup(ui_plugin_t ui)
{
    if (!ui->thin_interface) {
        fprintf(ui->fout,
		"%s: Terminal Mode Plugin cleaning up\r\n", ui->progname);
        fflush(ui->fout);
    }

    system("stty -raw echo");

    free(ui->title);
    free(ui->progname);
    free(ui);
}
