/* ui_plugin_wm_dockapp.c - implementation of a window-maker plugin for r2l. */

#include <dockapp.h>
#include <unistd.h>
#include <string.h>

#include "r2llib.h"
#include "ui_plugin.h"

/* pixmaps used for showing r2l's state and biditext's base state. */
#include "ln.xpm"
#include "ll.xpm"
#include "lr.xpm"
#include "rn.xpm"
#include "rl.xpm"
#include "rr.xpm"

/* parameters for initialization of a 'dockapp'. */
#define DESCRIPTION "\nA wm dockapp utility to control biditext.\n"
#define VERSION "r2ldockapp " PACKAGE_VERSION

/* size of window to ask 'dockapp' to allocate for us. */
#define WIN_SZ 56

/* mouse button codes, for the button press callback. */
#define BUTTON_LEFT   1
#define BUTTON_MIDDLE 2
#define BUTTON_RIGHT  3

/* plugin's data structure. */
struct ui_plugin_data {
    char* progname;	 	/* name of the program we're pluged into. */
    char* title;	 	/* title for the applet.                  */
    int thin_interface;	 	/* should we use a thin interface or not? */
    r2llib_t r2l_ctrl;   	/* r2l controller.                        */
    R2L_STATE md;	 	/* last known r2l mode.                   */
    BIDITEXT_BASE_STATE bs;	/* last known r2l mode.                   */
    void (*timer_cb)();  	/* application's timer callback.          */
    void *timer_cb_data; 	/* data for application's timer callback. */
    Pixmap pixmap_ln, mask_ln; 	/* r2l's state pixmaps and pixmap masks.  */
    Pixmap pixmap_ll, mask_ll;
    Pixmap pixmap_lr, mask_lr;
    Pixmap pixmap_rn, mask_rn;
    Pixmap pixmap_rl, mask_rl;
    Pixmap pixmap_rr, mask_rr;
};

/* since libdockapp's callbacks get no user parameters, */
/* we store the ui plugin's data in a global variable.  */
ui_plugin_t theUi = NULL;


/* callback invoked when the user presses any mouse button over r2l. */
void
buttonPressCallback(int button, int state, int x, int y)
{
    ui_plugin_t ui = theUi;
    int rc;

    if (button == BUTTON_LEFT){
	rc = r2l_disable(ui->r2l_ctrl);
	if (rc == 1) {
	    ui->md = R2L_DISABLED;
	    if (ui->bs == BT_BASE_NEUTRAL) {
    	        DASetPixmap(ui->pixmap_ln);
	    }
	    else if (ui->bs == BT_BASE_LTR) {
    	        DASetPixmap(ui->pixmap_ll);
	    }
	    else if (ui->bs == BT_BASE_RTL) {
    	        DASetPixmap(ui->pixmap_lr);
	    }
	}
    }
    else if (button == BUTTON_RIGHT){
	rc = r2l_enable(ui->r2l_ctrl);
	if (rc == 1) {
	    ui->md = R2L_ENABLED;
	    /* perform biditext base state checking. */
	    if (ui->bs == BT_BASE_LTR) {
    	        DASetPixmap(ui->pixmap_rl);
	    }
	    else if (ui->bs == BT_BASE_NEUTRAL) {
    	        DASetPixmap(ui->pixmap_rn);
	    }
	    else if (ui->bs == BT_BASE_RTL) {
    	        DASetPixmap(ui->pixmap_rr);
	    }
	}
    }
    else if (button == BUTTON_MIDDLE){
	/* perform biditext base state toggling. */
	if (ui->bs == BT_BASE_LTR) {
	    rc = r2l_set_biditext_base_state_neutral(ui->r2l_ctrl);
	    if (rc == 1) {
	        ui->bs = BT_BASE_NEUTRAL;
		if (ui->md == R2L_DISABLED) {
    	            DASetPixmap(ui->pixmap_ln);
		}
		else {
    	            DASetPixmap(ui->pixmap_rn);
		}
	    }
	}
	else if (ui->bs == BT_BASE_NEUTRAL) {
	    rc = r2l_set_biditext_base_state_rtl(ui->r2l_ctrl);
	    if (rc == 1) {
	        ui->bs = BT_BASE_RTL;
		if (ui->md == R2L_DISABLED) {
    	            DASetPixmap(ui->pixmap_lr);
		}
		else {
    	            DASetPixmap(ui->pixmap_rr);
		}
	    }
	}
	else if (ui->bs == BT_BASE_RTL) {
	    rc = r2l_set_biditext_base_state_ltr(ui->r2l_ctrl);
	    if (rc == 1) {
	        ui->bs = BT_BASE_LTR;
		if (ui->md == R2L_DISABLED) {
    	            DASetPixmap(ui->pixmap_ll);
		}
		else {
    	            DASetPixmap(ui->pixmap_rl);
		}
	    }
	}
    }
}

/* callback invoked by libdockapp's main loop, when its timer sets off. */
void
timerExpiredCallback()
{
    ui_plugin_t ui = theUi;

    (*ui->timer_cb)(ui->timer_cb_data);
}

/* initialize a given ui. returns a ui_plugin_t opaque pointer. */
ui_plugin_t
ui_init(int argc,
	char* argv[],
	r2llib_t r2l_ctrl,
	char* title,
	int thin_interface)
{
    unsigned int height, width;		/* for handling pixmaps.          */
    unsigned int x, y;			/* for drawing strings.           */
    int rc;
    DACallbacks callbacks = { NULL,	/* libdockapp's callbacks struct. */
			      buttonPressCallback,
			      NULL,
			      NULL,
			      NULL,
			      NULL,
			      timerExpiredCallback }; 
    char *displayName = "";		/* X display name.                */
    DAProgramOption options[] = {	/* libdockapp's cmd-line options. */
	{"-display", NULL, "Display to use", DOString, False, {&displayName}}
    };
    ui_plugin_t ui = (ui_plugin_t)malloc(sizeof(struct ui_plugin_data));

    theUi = ui;

    if (!ui)
	goto ui_ret;
	
    ui->r2l_ctrl = r2l_ctrl;
    ui->timer_cb = NULL;
    ui->timer_cb_data = NULL;
    ui->md = R2L_ERROR;
    ui->bs = BT_BASE_ERROR;
    ui->thin_interface = thin_interface;
    ui->title = strdup(title);
    if (!ui->title)
	goto err_free_ui;

    DAParseArguments(argc,
		     argv,
		     options,
		     sizeof (options) / sizeof (DAProgramOption),
		     DESCRIPTION,
		     VERSION);
    DAInitialize(displayName, "r2l", WIN_SZ, WIN_SZ, argc, argv);

    DAMakePixmapFromData(ln_xpm, &ui->pixmap_ln, &ui->mask_ln, &height, &width);
    DAMakePixmapFromData(ll_xpm, &ui->pixmap_ll, &ui->mask_ll, &height, &width);
    DAMakePixmapFromData(lr_xpm, &ui->pixmap_lr, &ui->mask_lr, &height, &width);
    DAMakePixmapFromData(rn_xpm, &ui->pixmap_rn, &ui->mask_rn, &height, &width);
    DAMakePixmapFromData(rl_xpm, &ui->pixmap_rl, &ui->mask_rl, &height, &width);
    DAMakePixmapFromData(rr_xpm, &ui->pixmap_rr, &ui->mask_rr, &height, &width);

    /* set the title on the loaded pixmaps. */
    {
	XFontStruct* font_info = NULL;
	int font_height;

	char* font_pattern = "*-fixed-*-8-*";

	font_info = XLoadQueryFont(DADisplay, font_pattern);
	if (!font_info) {
	    printf("XLoadQueryFont failed\n");
	    exit(1);
	}
        rc = XSetFont(DADisplay, DAGC, font_info->fid);

	XSetForeground(DADisplay, DAGC, WhitePixel(DADisplay, DefaultScreen(DADisplay)));
	XFillRectangle(DADisplay, ui->pixmap_ln, DAGC, 0, 0, 24, 8);
	XFillRectangle(DADisplay, ui->pixmap_ll, DAGC, 0, 0, 24, 8);
	XFillRectangle(DADisplay, ui->pixmap_lr, DAGC, 0, 0, 24, 8);
	XFillRectangle(DADisplay, ui->pixmap_rn, DAGC, 0, 0, 24, 8);
	XFillRectangle(DADisplay, ui->pixmap_rl, DAGC, 0, 0, 24, 8);
	XFillRectangle(DADisplay, ui->pixmap_rr, DAGC, 0, 0, 24, 8);
	XSetForeground(DADisplay, DAGC, BlackPixel(DADisplay, DefaultScreen(DADisplay)));

	font_height = font_info->ascent + font_info->descent;
        x = 1;
	y = font_height;
        XDrawString(DADisplay, ui->pixmap_ln, DAGC, x, y, ui->title, strlen(ui->title));
        XDrawString(DADisplay, ui->pixmap_ll, DAGC, x, y, ui->title, strlen(ui->title));
        XDrawString(DADisplay, ui->pixmap_lr, DAGC, x, y, ui->title, strlen(ui->title));
        XDrawString(DADisplay, ui->pixmap_rn, DAGC, x, y, ui->title, strlen(ui->title));
        XDrawString(DADisplay, ui->pixmap_rl, DAGC, x, y, ui->title, strlen(ui->title));
        XDrawString(DADisplay, ui->pixmap_rr, DAGC, x, y, ui->title, strlen(ui->title));
    }

    DASetPixmap(ui->pixmap_ln);
    DASetCallbacks(&callbacks);
    DAShow();

    goto ui_ret;

err_free_ui:
    free(ui);
    ui = NULL;

ui_ret:
    return ui;
}

/* define whether the ui plugin runs its own main loop. */
/* returns '1' for true, '0' for false.                 */
/* note that if it does, it must also support the timer */
/* registration interface.                              */
int
ui_provides_main_loop(ui_plugin_t ui)
{
    return 1;
}

/* execute the ui plugin's main loop - invoked only if */
/* 'ui_provides_main_loop' returned '1' previously.    */
void
ui_main_loop(ui_plugin_t ui)
{
    DAEventLoop();
}

/* register a callback with the ui plugin to be invoked */
/* every 'ms' milliseconds.                             */
/* invoked only if 'ui_provides_main_loop' returned '1' */
/* previously.                                          */
int ui_register_timer_callback(ui_plugin_t ui,
			       int ms,
			       int (*cb)(void*),
			       void* data)
{
    ui->timer_cb = (void(*)())cb;
    ui->timer_cb_data = data;
    DASetTimeout(1000);

    return 1;
}

/* notify the ui plugin about a mode change. */
void
ui_notify_of_r2l_mode(ui_plugin_t ui, R2L_STATE md, BIDITEXT_BASE_STATE bs)
{
    int was_change = 0;
    if (md != ui->md) {
	ui->md = md;
	was_change = 1;
    }
    if (bs != ui->bs) {
        ui->bs = bs;
        was_change = 1;
    }
    if (was_change) {
	if (md == R2L_DISABLED) {
    	    DASetPixmap(bs == BT_BASE_NEUTRAL ? ui->pixmap_ln :
			    (bs == BT_BASE_LTR ? ui->pixmap_ll :
			        ui->pixmap_lr));
	}
	else {
    	    DASetPixmap(bs == BT_BASE_NEUTRAL ? ui->pixmap_rn :
			    (bs == BT_BASE_LTR ? ui->pixmap_rl :
			        ui->pixmap_rr));
	}
    }
}

/* returns a file descriptor that the UI would like r2l to listen on.  */
/* we run our own main loop - so we return an invalid descriptor here. */
int
ui_get_listening_fd(ui_plugin_t ui)
{
    /* do nothing - it won't be invoked. */
    return -1;
}

/* callback invoked when r2l discovers the file descriptor of the UI is active. */
int
ui_perform_callback(ui_plugin_t ui, r2llib_t r2l_ctrl)
{
    /* do nothing. */
    return 0;
}

/* cleanup the ui plugin. after this call, 'ui' may no longer be used. */
void
ui_cleanup(ui_plugin_t ui)
{
    /* do nothing. */
}
