/*
 * $Id: r2l.c,v 1.3 2002/01/23 02:44:48 tzafrir Exp $
 */
/* r2l.c: A simple command-line driver for the r2l library */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>

#include "r2llib.h"

/* I want this to have a different error code than libr2l's */
#define INCORRECT_ARGUMENT (2)
#define LIBR2L_FAILED (1)
#define NULL_STRING "[NULL]"

/** enable/disable debug mode */
static int DebugFlag = 0;

/** hold the name of base directionality state */
struct {
     BIDITEXT_BASE_STATE state;
     const char* name;
} BiditextBaseStateNames[] = {
     {BT_BASE_NEUTRAL,"neutral"}, /* neutral = 0*/
     {BT_BASE_LTR    ,"ltr"},
     {BT_BASE_RTL    ,"rtl"},
     {BT_BASE_ERROR  ,"ERROR"}
};

/** Hold the name of the r2l enabled/disabled state */
struct {
     R2L_STATE state;
     const char* name;
} BiditextStateNames[] = {
     {R2L_ENABLED ,"enabled"}, /* enabled = 0 */
     {R2L_DISABLED,"disabled"},
     {R2L_ERROR   ,"ERROR"}
};

/** hold the names of r2llib error types */
struct {
     R2L_ERROR_TYPE type;
     const char* name;
} R2lErrorTypeNames[] = {
     {R2L_SUCCESS,          "success"},      /* R2L_SUCCESS = 0 */
     {R2L_ERR_OOM,          "out of memory"},
     {R2L_ERR_INVALID_REPR, "invalid representation"},
     {R2L_ERR_ACCESS,       "file access error"}
};

/** prints a trace message when in debug mode */
void print_error(const char* message){
     printf("\nr2l error:%s\n\n", message);
}

/** @param error_message an error message string, or NULL for a usage message */
void print_usage_message(const char* error_message) {
     if (error_message != NULL) {
	  print_error(error_message);
     }
     printf(
	  "r2l: a command-line interface to r2llib\n"
	  "Usage:\n"
	  "  r2l p [<enabled|base|token|state|all>]\n"
	  "  r2l s [<enable|disable|toggle>]\n"
	  "  r2l b [<neutral|ltr|rtl>]\n"
	  "\n"
	  "    p: Print information (Default: all)\n"
	  "    s: set State: enable/disable/toggle biditext (default:enable)\n"
	  "    b: set Base direction: neutral/rtl/ltr (default: neutral)\n"
	  );
}

/** prints a trace message when in debug mode */
void print_trace(const char* message){
     if (DebugFlag){
	  printf("r2l trace:%s\n", message);
     }
}


/** Currently everything is done inside main() */
int 
main(int argc, char* argv[])
{
     R2L_ERROR_TYPE err;
     int ret_val = 0; /* default to success */
     int r2l_ret_val = -1; /* default to failure */
     /* Already processes the --file-name argument */
     r2llib_t rtl = r2l_init(&argc, argv, &err);
     if (!rtl) {
	  char buff[1024];
	  sprintf(buff,"r2llib initialization failed: %s", 
		  R2lErrorTypeNames[err].name);
	  print_error(buff);
	  exit(LIBR2L_FAILED);
     }

     if (argc < 2 || argv[1] == NULL) {
	  print_usage_message("Command missing");
	  r2l_destroy(rtl);
	  exit(INCORRECT_ARGUMENT);
     }

     switch (argv[1][0]) {
     case 'p': /* print */
	  {
	       BIDITEXT_BASE_STATE base;
	       R2L_STATE state;
	       const char* token;

	       r2l_query_state(rtl, &state, &base);
	       if ((state == R2L_ERROR)||(base == BT_BASE_ERROR)) {
		    ret_val = LIBR2L_FAILED;
	       /* default value */
	       } else if (((argc < 3) || (argv[2] == NULL)) ||
	                 (!strcasecmp(argv[2],"all"))) {
		    token = r2l_get_text_serialization(rtl);
		    if (token == NULL) {token = NULL_STRING;}
		    printf ("State: %s\nBase direction: %s\nToken: %s\n", 
			    BiditextStateNames[state].name, 
			    BiditextBaseStateNames[base].name,
			    token);
	       } else if (!strcasecmp(argv[2],"enabled")) {
		    printf ("%s\n", (BiditextStateNames[state]).name);
	       } else if (!strcasecmp(argv[2],"state")) {
		    printf ("State: %s\nBase direction: %s\n", 
			    BiditextStateNames[state].name, BiditextBaseStateNames[base].name);
	       } else if (!strcasecmp(argv[2],"token")) {
		    token = r2l_get_text_serialization(rtl);
		    if (token == NULL) {token = NULL_STRING;}
		    printf("%s\n",token);
		    break;
	       } else if (!strcasecmp(argv[2],"base")) {
		    printf ("%s\n", (BiditextBaseStateNames[base]).name);
	       } else {
		    print_usage_message("Incorrect State setting command");
		    ret_val=INCORRECT_ARGUMENT;
	       }
	  }
	  break;
     case 's': /* change State */
	  if ((argc < 3 || argv[2] == NULL) || /* default value */
	            (!strcasecmp(argv[2],"enable"))) {
	       r2l_ret_val = r2l_enable(rtl);
	  } else if (!strcasecmp(argv[2],"disable")) {
	       r2l_ret_val = r2l_disable(rtl);
	  } else if (!strcasecmp(argv[2],"toggle")) {
	       r2l_ret_val = r2l_toggle(rtl);
	  } else {
	       print_usage_message("Incorrect State setting command");
	       ret_val = INCORRECT_ARGUMENT;
	       break;
	  }
	  if (r2l_ret_val != 1) {
	       print_error("r2llib failed to set the r2l state");
	       ret_val = LIBR2L_FAILED;
	       break;
	  }
	  break;
     case 'b': /* change Base direction */
	  if ((argc < 3 || argv[2] == NULL) || /* default value */
		    (!strcasecmp(argv[2],"neutral"))) {
	       r2l_ret_val = r2l_set_biditext_base_state_neutral(rtl);
	  } else if (!strcasecmp(argv[2],"rtl")) {
	       r2l_ret_val = r2l_set_biditext_base_state_rtl(rtl);
	  } else if (!strcasecmp(argv[2],"ltr")) {
	       r2l_ret_val = r2l_set_biditext_base_state_ltr(rtl);
	  } else {
	       print_usage_message("Incorrect Base direction setting command");
	       ret_val = INCORRECT_ARGUMENT;
	  }
	  if (r2l_ret_val != 1) {
	       print_error("r2llib failed to set the biditext base state");
	       ret_val = LIBR2L_FAILED;
	       break;
	  }
	  break;
     default: print_usage_message("Incorrect command");
	  ret_val = INCORRECT_ARGUMENT;
     }
     r2l_destroy(rtl);
     return ret_val;
}

