/*
 * $Id: r2llib.c,v 1.1.1.1 2001/09/13 12:42:16 mulix Exp $
 */
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <assert.h>

#include "r2llib.h"
#include "r2llib_data.h"
#include "parse_argv.h"
#include "r2l_state.h"
#include "biditext_base_state.h"
#include "query_state.h"
#include "access.h"

#define FILE_NAME_ENV_VARIABLE "BIDITEXT_FILENAME"
#define DEFAULT_FILE_NAME ".rev"


/*
 * initalize the r2l library.
 * parse argv and remove the following options (if they were given)
 * * --file-name [file-name]
 *  @returns a valid r2lib_t for use in subsequent calls to the r2l library.
 *  if an error occurs, fills err with error info, otherwise err is unchanged
 */
r2llib_t
r2l_init(int* argc, char* argv[], R2L_ERROR_TYPE* err)
{
     char* fname = NULL;
     char* prefix = NULL;
     int free_prefix = 0;
     char* tmpfname = NULL;
     r2llib_t rtl = NULL;
     R2L_ERROR_TYPE error = R2L_SUCCESS;

     /* try to get the file name from the command line */
     parse_command_line(argc, argv, &fname);

     /* no go? maybe from the environment? */
     if (!fname)
	  fname = getenv(FILE_NAME_ENV_VARIABLE);

     /* still no go? oh well, default it is */
     if (!fname){
	  fname = DEFAULT_FILE_NAME;
	  prefix = getenv("HOME");
     } else {
	  if (fname[0] != '/'){
	       /* we ignore paths longer than 1024 bytes, since i'm lazy */
	       prefix = getcwd(NULL, 1024);
	       free_prefix = 1;
	  }
     }

     if (prefix){
	  int len = strlen(prefix) + 1+ strlen(fname);
	  tmpfname = (char*)malloc(len + 1);
	  if (!tmpfname){
	       error = R2L_ERR_OOM;
	       goto done;
	  }
	  snprintf(tmpfname, len+1, "%s/%s", prefix, fname);
	  fname = tmpfname;
     }

     /* we got the file name, now check if we can access it */
     error = is_allowed_access(fname);
     if (error != R2L_SUCCESS)
	  goto done;

     /* finally, all is well, create the opaque token */
     rtl = create_r2llib_data(fname);

 done:
     if (tmpfname)
	  free(tmpfname);

     if (free_prefix && prefix)
	  free(prefix);

     if (err && (error != R2L_SUCCESS))
	  *err = error;

     return rtl;
}

/* perform cleanup */
void
r2l_destroy(r2llib_t rtl)
{
     assert(rtl);

     destroy_r2llib_data(rtl);
}

/* return 1 on success, -1 on failure */
int 
r2l_enable(r2llib_t rtl)
{
     assert(rtl);

     if (set_r2l_state(r2llib_data_get_file_name(rtl), R2L_ENABLED) 
	 != R2L_ENABLED)
	  return -1;

     return 1;
}


/* return 1 on success, -1 on failure */
int 
r2l_disable(r2llib_t rtl)
{
     assert(rtl);
     
     if (set_r2l_state(r2llib_data_get_file_name(rtl), R2L_DISABLED) 
	 != R2L_DISABLED)
	  return -1;
     
     return 1;
}

/* return 1 on success, -1 on failure */
int
r2l_toggle(r2llib_t rtl)
{
     assert(rtl);
     
     if (toggle_r2l_state(r2llib_data_get_file_name(rtl)) != R2L_ERROR)
	  return 1;

     return -1;
}

R2L_STATE
r2l_get_current_r2l_state(r2llib_t rtl)
{
     assert(rtl);
     
     return get_r2l_state(r2llib_data_get_file_name(rtl));
}

int 
r2l_set_biditext_base_state_neutral(r2llib_t rtl)
{
     assert(rtl);
     
     if (set_biditext_base_state(r2llib_data_get_file_name(rtl), BT_BASE_NEUTRAL) 
	 != BT_BASE_NEUTRAL)
	  return -1;
     
     return 1;
}

int
r2l_set_biditext_base_state_rtl(r2llib_t rtl)
{
     assert(rtl);

     if (set_biditext_base_state(r2llib_data_get_file_name(rtl), BT_BASE_RTL) 
	 != BT_BASE_RTL)
	  return -1;
     
     return 1;
}

int 
r2l_set_biditext_base_state_ltr(r2llib_t rtl)
{
     if (set_biditext_base_state(r2llib_data_get_file_name(rtl), BT_BASE_LTR) 
	 != BT_BASE_LTR)
	  return -1;
     
     return 1;
}

BIDITEXT_BASE_STATE
r2l_get_current_biditext_base_state(r2llib_t rtl)
{
     assert(rtl);

     return get_biditext_base_state(r2llib_data_get_file_name(rtl));
}

const char* 
r2l_get_text_serialization(r2llib_t rtl)
{
     assert(rtl);
     
     return r2llib_data_get_file_name(rtl);
}

/** @return a new r2l token from this textual representation */
r2llib_t 
r2l_init_from_text_serialization(const char* repr, R2L_ERROR_TYPE* err)
{
     r2llib_t tok; 
     R2L_ERROR_TYPE error;
     
     if (!repr)
	  return r2l_init(NULL, NULL, err); 

     error = is_allowed_access(repr);
     if (error != R2L_SUCCESS){
	  if (err)
	       *err = error;
	  return NULL;
     }

     tok = create_r2llib_data(repr);
     if (err && !tok)
	  *err = R2L_ERR_OOM;

     return tok;
}

void
r2l_query_state(r2llib_t rtl, R2L_STATE* r2l_state, 
		BIDITEXT_BASE_STATE* bt_state)
{
     const char* file_name; 

     assert(rtl);

     file_name = r2llib_data_get_file_name(rtl);
     if (file_name){
	  r2l_query_state_internal(file_name, r2l_state, bt_state);
     }
}

void 
r2l_set_state(r2llib_t rtl, 
	      R2L_STATE* r2l_state,
	      BIDITEXT_BASE_STATE* bt_state)
{
     const char* file_name;
     R2L_STATE curst;
     BIDITEXT_BASE_STATE curbt;
     

     assert(rtl); 

     file_name = r2llib_data_get_file_name(rtl);

     if (!r2l_state || !bt_state){
	  /* if we didn't get st or bt, put a place holder for it, 
	     since r2l_set_state_internal expects both */

	  r2l_query_state_internal(file_name, &curst, &curbt);
	  if (!r2l_state)
	       r2l_state = &curst;

	  if (!bt_state)
	       bt_state = &curbt;
     }

     r2l_set_state_internal(file_name, r2l_state, bt_state);  
}


