#include <map>
#include <assert.h>
#include "auto_display.h"
#include "debug.h"

using namespace std;

class AutoDisplayImpl : public AutoDisplay {
public:
  explicit AutoDisplayImpl(const string& dispName);
  static AutoDisplay* newInstance(const string& dispName);
  virtual void release(void);
  virtual Display* getDisplay(void);
  virtual const Display* getDisplay(void) const;


private:
  Display* dpy;
  int refCount;
  const string dpyName;
  static map<string,AutoDisplayImpl*> m_displays;
  static const char* getCStrDpyName(const std::string& dpyName);
  bool isValid(void) const;

    //Disallowed methods
private:
  AutoDisplayImpl(const AutoDisplayImpl&);
  AutoDisplayImpl& operator = (const AutoDisplayImpl&);
};



AutoDisplayImpl::AutoDisplayImpl(const string& dispName)
  :dpy(XOpenDisplay(getCStrDpyName(dispName))),
   refCount(1),
   dpyName(dispName)
{
  //XOpenDisplay might fail. Use isValid to check!!
  if (!isValid()) {
    WARNING("%s:%d Warning, cannot open display %s\n",
	    __FILE__,__LINE__,dispName.c_str());
  }

}
 

bool AutoDisplayImpl::isValid(void) const
{
  return dpy != NULL;
}


AutoDisplay* AutoDisplayImpl::newInstance(const string& dispName)
{
  map<string,AutoDisplayImpl*>::const_iterator it = m_displays.find(dispName);

  if (it != m_displays.end()) {
    MESSAGE("%s:%d Info: AutoDisplayImpl::newInstance() display '%s' "
	    "already opened\n",__FILE__,__LINE__,dispName.c_str());
    (*it).second->refCount++;
    return (*it).second;
  }

  AutoDisplayImpl* newDisplay = new AutoDisplayImpl(dispName);

  if  (newDisplay == NULL) {
    return NULL;
  }

  if (!newDisplay->isValid()) {
    newDisplay->release();
    return NULL;
  }

  m_displays[dispName] = newDisplay;
  
  MESSAGE("%s:%d Info: Returning a new valid instance for display '%s'\n",
	  __FILE__,__LINE__,dispName.c_str());
  return newDisplay;
  
}


void AutoDisplayImpl::release(void)
{
  MESSAGE("%s:%d Info: AutoDisplay::release() refCount = %d\n",
	  __FILE__,__LINE__,refCount);
  if (--refCount == 0) {
    if (isValid()) {
      m_displays.erase(dpyName);
      MESSAGE("%s:%d Info: Closing connection to to display '%s'\n",
	      __FILE__,__LINE__,dpyName.c_str());
      XCloseDisplay(dpy);
    }
    delete this;
  }
}


Display* AutoDisplayImpl::getDisplay(void)
{
  return dpy;
}

const Display* AutoDisplayImpl::getDisplay(void) const
{
  return dpy;
}


const char* AutoDisplayImpl::getCStrDpyName(const std::string& dpyName)
{
  if (dpyName == "")
    return NULL;

  return dpyName.c_str();
}
 

map<string,AutoDisplayImpl*> AutoDisplayImpl::m_displays;


AutoDisplay* AutoDisplay:: getInstance(const string& dispName)
{
  return AutoDisplayImpl::newInstance(dispName);
}


AutoDisplay::AutoDisplay(void) {}


AutoDisplay::~AutoDisplay(void) {}


bool AutoDisplayCompare::operator()(const AutoDisplay* ad1,
				    const AutoDisplay* ad2) const
{
  return ad1->getDisplay() < ad2->getDisplay();
}
