#include <assert.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "communication.h"
#include "refresher.h"
#include "messages.h"
#include "window_db_iterator.h"
#include "debug.h"
#include "refreshd_params.h"

extern WindowDB winDB;


#if NOISE_LEVEL == NOISE_LEVEL_MAX

static void dumpMessage(const Message* msg)
{
  switch (msg->type) {
  case REFRESHD_PING: {
    MESSAGE("type: REFRESHD_PING"
	    " pid: %ld",
	    (long)msg->pingMsg.pid);
  }
  break;
  
  case REFRESHD_REPLY_PING: {
    MESSAGE("type: REFRESHD_REPLY_PING"
	    " pid: %ld",
	    (long)msg->replyPingMsg.pid);
  }
  break;

  case REFRESHD_ADD_WINDOW: {
    const AddWindowMessageData* awmd = &msg->addWindowMsg;
    MESSAGE("type: REFRESHD_ADD_WINDOW"
	    " dpy: %s" 
	    " w: 0x%lx" 
	    " bidiTextKeyAtom: 0x%lx"
	    " pid: %ld",
	    awmd->dispName,
	    (unsigned long)awmd->w,
	    (unsigned long)awmd->bidiTextKeyAtom,
	    (long)awmd->pid);
  }
  break;
  
  case REFRESHD_REPLY_ADD_WINDOW: {
    MESSAGE("type: REFRESHD_REPLY_ADD_WINDOW"
	    " pid: %ld",
	    (long)msg->replyAddWindowMsg.pid);
  }
  break;
  
  case REFRESHD_REFRESH_NOW:
    MESSAGE("type: REFRESHD_RFRESH_NOW");
    break;
    
  default:
    WARNING("Warning: unknown message type (%d)",(int)msg->type);
    break;
  }
}

#endif /* NOISE_LEVEL == NOISE_LEVEL_MAX */



void* communicator(void* arg)
{
  int sd;
  Message msg;
  Message rply;
  struct sockaddr_un sa;
  char* addr_ptr = sa.sun_path;

  sd = socket(PF_UNIX,SOCK_DGRAM,0);
  if (sd < 0) {
     ERROR("%s:%d Fatal error: Cannot create socket. %m "
	   "Refresh daemon dying\n",
	  __FILE__,__LINE__);
    close(sd);
    exit(1);
  }

  memset(&sa,0,sizeof(sa));

  sa.sun_family = AF_UNIX;

#ifdef ABSTRACT_UNIX_SOCKET_ADDRESSES
  *addr_ptr++ = '\0';
#endif

  //Just play safe here and ensure that the address will be NUL teminated
  strncpy(addr_ptr,REFRESHD_SOCKET_NAME,sizeof(sa.sun_path) - 2);

  int rc = bind(sd,(struct sockaddr*)&sa,sizeof(sa));

  if (rc < 0) {
    ERROR("%s:%d Fatal error: Cannot bind socket. %m Refresh daemon dying\n",
	  __FILE__,__LINE__);
    close(sd);
    exit(1);
  }

  for (;;) {
    struct sockaddr_un from;
    size_t fromlen = sizeof(from);

    int numRecvRetries = 0;

  recv_again:
    rc = recvfrom(sd,&msg,sizeof(msg),0,(struct sockaddr*)&from,&fromlen);

    if (rc < 0) {
      if (numRecvRetries++ < MAX_RECV_RETRIES) {
	goto recv_again;
      }
      ERROR("%s:%d Fatal error: Cannot receive message. %m "
	    "Refresh daemon dying\n",
	    __FILE__,__LINE__);
      close(sd);
#ifndef ABSTRACT_UNIX_SOCKET_ADDRESSES
    unlink(REFRESHD_SOCKET_NAME);
#endif
      exit(1);
    }

    if (rc != sizeof(msg)) {
      WARNING("%s:%d Illegal message size: (%d). Message ignored\n",
	      __FILE__,__LINE__,rc);
      continue;
    }

#if NOISE_LEVEL == NOISE_LEVEL_MAX
    MESSAGE("%s:%d Received ",__FILE__,__LINE__);
    dumpMessage(&msg);
    MESSAGE("\n");
#endif

    switch(msg.type) {
    case REFRESHD_ADD_WINDOW: {
      AddWindowMessageData* awmd = &msg.addWindowMsg;
      awmd->dispName[MAX_DISP_NAME-1] = '\0'; //paranoia

      AutoDisplay* adpy = AutoDisplay::getInstance(awmd->dispName);

      if (adpy == NULL) {
	WARNING("%s:%d Warning: Cannot open display '%s'"
		"Message ignored\n",
		__FILE__,__LINE__,awmd->dispName);
	continue;
      }

      const char* bidiTextKey = XGetAtomName(adpy->getDisplay(),
					     awmd->bidiTextKeyAtom);

      if (bidiTextKey == NULL) {
	WARNING("%s:%d Warning: Invalid biditext key atom: 0x%lx"
		"Message ignored\n",
		__FILE__,__LINE__,awmd->bidiTextKeyAtom);
	adpy->release();
	continue;
      }

      winDB.lock();
      winDB.add(awmd->dispName,
		awmd->w,
		bidiTextKey);
      winDB.unlock();

      adpy->release();

      rply.type = REFRESHD_REPLY_ADD_WINDOW;
      rply.replyAddWindowMsg.pid = awmd->pid;

      int numSendRetries = 0;
    send_add_window_reply_again:
      int rc = sendto(sd,&rply,sizeof(rply),0,
		      (struct sockaddr*)&from,
		      sizeof(from));
      if (rc < 0) {
	if (numSendRetries++ <  MAX_SEND_RETRIES)
	  goto  send_add_window_reply_again;
	WARNING("%s:%d Warning: Cannot send reply message. %m\n",
		__FILE__,__LINE__);
      }
    }
    break;

    case REFRESHD_PING: {
      PingMessageData* pmd = &msg.pingMsg;

      rply.type = REFRESHD_REPLY_PING;
      rply.replyPingMsg.pid = pmd->pid;

      int numSendRetries = 0;
    send_ping_reply_again:
      int rc = sendto(sd,&rply,sizeof(rply),0,
		      (struct sockaddr*)&from,
		      sizeof(from));
      if (rc < 0) {
	if (numSendRetries++ <  MAX_SEND_RETRIES)
	  goto  send_ping_reply_again;
	WARNING("%s:%d Warning: Cannot send reply message. %m\n",
		__FILE__,__LINE__);
      }
    }
    break;

    case REFRESHD_REFRESH_NOW:
      refreshNow();
      break;
      
    default:
      WARNING("%s:%d Warning, received illegal message type: %d\n",
	      __FILE__,__LINE__,(int)msg.type);
      break;
    }
  }
}
