#include <assert.h>
#include <iostream>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>

#include <pthread.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifndef ABSTRACT_UNIX_SOCKET_ADDRESSES
#include <sys/stat.h>
#endif

#include "window_db.h"
#include "error_handler.h"
#include "communication.h"
#include "refresher.h"
#include "scavenger.h"
#include "debug.h"
#include "refreshd_params.h"
#include "refreshd_cnt.h"

using namespace std;

WindowDB winDB;

static bool noAuto;
static pid_t childPid;


static void usage(int argc,char* argv[])
{
  cerr << "Usage: " << argv[0] << " [--no_auto]" << endl;
}

static void signalHandler(int sig)
{
  MESSAGE("Have a nice death, my son (or daughter, or whatever you are :-)\n");
  kill(childPid,sig);
}





static int childMain(void)
{
  int rc;
  pthread_t auto_refresher_tid;
  pthread_t communicator_tid;
  pthread_t scavenger_tid;
  Status mtXlib;
  const char* errMsg = "No error";

  mtXlib = XInitThreads();
    
  if (!mtXlib) {
    errMsg = "Xlib is not thread safe. refreshd cannot run";
    goto err_mt_xlib;
  }

  initXErrorHandler();


  rc = pthread_create(&communicator_tid,NULL,
		      communicator,NULL);
  
  if (rc != 0) {
    errMsg = "Cannot create communicator thread. %m";
    goto err_create_communicator;
  }
  
  if (!noAuto) {
    rc = pthread_create(&auto_refresher_tid,NULL,
			autoRefresher,NULL);
    if (rc != 0) {
      errMsg = "Cannot create auto-refresher thread. %m";
      goto err_create_auto_refresher;
    }
  }
  
  
  rc = pthread_create(&scavenger_tid,NULL,
		      scavenger,NULL);
  
  if (rc != 0) {
    errMsg = "Cannot create scavenger thread. %m";
    goto err_create_scavenger;
  }
  
  /* let the main thread become the refresher */
  refresher(NULL);

  return 0;
  
 err_create_scavenger:
 err_create_auto_refresher:
  
#ifndef ABSTRACT_UNIX_SOCKET_ADDRESSES
  unlink(REFRESHD_SOCKET_NAME);
#endif
  
 err_create_communicator:
 err_mt_xlib:
  ERROR(errMsg);
  return 1;
}



int main(int argc, char* argv[])
{
  const char* errMsg = "No error";
  RD_CNT_STATUS status;
#ifndef ABSTRACT_UNIX_SOCKET_ADDRESSES
  struct stat sbuff;
  int rc;
#endif

  switch (argc) {
  case 1:
    break;
  case 2:
    if (strcmp(argv[1],"--no_auto") == 0)
      noAuto = true;
    else {
      usage(argc,argv);
      exit(1);
    }
      
    break;

  default:
    usage(argc,argv);
    exit(1);
  }

// Solaris does not has LOG_PERROR_ defined. In this case: ignore this option
#ifndef LOG_PERROR
#  define LOG_PERROR 0
#endif

  openlog("refreshd",LOG_PERROR,LOG_DAEMON);

  /* Check if another intance is not already running */
  status = rdCntInitialize();

  if (status == RD_CNT_SUCCESS) {
    errMsg = "It seems that another instance of refreshd is already running\n"
      "Kill it first with a TERM or INT signal and try again";
    goto err_already_running;
  }

#ifndef ABSTRACT_UNIX_SOCKET_ADDRESSES
  rc = stat(REFRESHD_SOCKET_NAME,&sbuff);
  if (rc == 0) {
    errMsg = "A file named " REFRESHD_SOCKET_NAME
      " already exists. Most likely it is a leftover from an abnormally "
      "teminated refresh daemon. Delete the file and try again.";
    goto err_socket_exists;
  }
#endif


  childPid = fork();

  if (childPid < 0) {
    errMsg = "Cannot fork. %m";
    goto err_fork;
  }
  else if (childPid == 0) {
    int rc = childMain();
    /* If child_main returns, this is surely because of an error */
    exit(rc);
  }
  else {
    int status;
    int rc;

    signal(SIGTERM,signalHandler);
    signal(SIGINT,signalHandler);
  again:
    rc = wait(&status);
    if (rc < 0 && errno == EINTR)
      goto again;
#ifndef ABSTRACT_UNIX_SOCKET_ADDRESSES
    /* Clean up the child excements */
    unlink(REFRESHD_SOCKET_NAME);
#endif
    closelog();
    exit(0);
  }

#ifndef ABSTRACT_UNIX_SOCKET_ADDRESSES
 err_socket_exists:
#endif
 err_already_running:
 err_fork:
  ERROR(errMsg);
  closelog();
  return 1;
}
