#include <assert.h>

#include "window_db.h"
#include "window_db_iterator.h"
#include "debug.h"

using namespace std;

WindowDB::WindowDB(void)
  :m_isLocked(false)
{
  int status = pthread_mutex_init(&m_dbLock,NULL);
  assert(status == 0);
}


WindowDB::~WindowDB(void)
{
  pthread_mutex_destroy(&m_dbLock);
}


WindowDB::Result  WindowDB::add(const string& dispName,
				Window w,
				const string& bidiTextKey)
{
  assert(m_isLocked);

  if (!m_isLocked) {
    ERROR("%s:%d WindowDB bug: database not locked\n",
	  __FILE__,__LINE__);
    return FAILURE;
  }
  //use the side effect of operator []
  struct PerBidiTextKeyData& bidiTextKeyData = m_bidiTextKeys[bidiTextKey];

  if (bidiTextKeyData.r2llib == NULL) {
    R2L_ERROR_TYPE err;

    bidiTextKeyData.r2llib = 
      r2l_init_from_text_serialization(bidiTextKey.c_str(),
				       &err);

    if (bidiTextKeyData.r2llib == NULL) {
      m_bidiTextKeys.erase(bidiTextKey);


      WARNING("%s:%d Warning: r2l_init_from_text_serialization() "
	      "failed for bidiTextKey '%s' (error code: %d)\n",
	      __FILE__,__LINE__,bidiTextKey.c_str(),(int)err);

      return FAILURE;
    }
  }

  DisplayMap& dpyMap = bidiTextKeyData.dpyMap;

  AutoDisplay* adpy = AutoDisplay::getInstance(dispName);

  if (adpy == NULL) {

    WARNING("%s:%d Warning: Cannot open display %s\n",
	    __FILE__,__LINE__,dispName.c_str());

    if (dpyMap.empty())
      m_bidiTextKeys.erase(bidiTextKey);
    
    return FAILURE;
  }
  
  WindowSet& wnds = dpyMap[adpy];

  if (!wnds.empty()) 
    adpy->release();

  wnds.insert(w);

  return SUCCESS;
}


WindowDB::Result  WindowDB::remove(BidiTextKeyIterator& bidiTextKeyIt,
				   DisplayMapIterator& dpyIt,
				   Window w)
{
  assert(m_isLocked);
  
  if (!m_isLocked) {
    ERROR("%s:%d WindowDB bug: database not locked\n",
	  __FILE__,__LINE__);
    return FAILURE;
  }

  if (bidiTextKeyIt == m_bidiTextKeys.end()) {
    ERROR("%s:%d Error: bidiTextKeyIt points to end\n",
	  __FILE__,__LINE__);
    return FAILURE;
  }
  
  PerBidiTextKeyData& pbtkd = (*bidiTextKeyIt.m_it).second;
  r2llib_t& r2llib = pbtkd.r2llib;
  DisplayMap& dpyMap = pbtkd.dpyMap;

  if (dpyIt == dpyMap.end()) {
     ERROR("%s:%d Error: dpyIt points to end\n",
	   __FILE__,__LINE__);
    return FAILURE;
  }

  WindowSet& wnds = (*dpyIt).second;
  
  WindowSet::iterator wndsIt = wnds.find(w);

  if (wndsIt == wnds.end()) {
    ERROR("%s:%d Error: wndsIt points to end\n",
	  __FILE__,__LINE__);
    return FAILURE;
  }

  MESSAGE("%s:%d Info: Removing window id %lx\n",
	  __FILE__,__LINE__,(unsigned long)w);
  wnds.erase(wndsIt);

  if (wnds.empty()) {
    MESSAGE("%s:%d Info: Removing display '%s'\n",
	    __FILE__,__LINE__,DisplayString((*dpyIt).first->getDisplay()));
    (*dpyIt).first->release();
    dpyMap.erase(dpyIt);
  }

  if (dpyMap.empty()) {
    MESSAGE("%s:%d Info: Removing biditext key '%s'\n",
	    __FILE__,__LINE__,(*bidiTextKeyIt.m_it).first.c_str());
    r2l_destroy(r2llib);
    r2llib = NULL;
    m_bidiTextKeys.erase(bidiTextKeyIt.m_it);
  }

  return SUCCESS;
}



void WindowDB::lock(void)
{
  pthread_mutex_lock(&m_dbLock);
  m_isLocked = true;
}

void WindowDB::unlock(void)
{
  assert(m_isLocked);
  if (!m_isLocked) {
    ERROR("%s:%d WindowDB Bug: database is not locked");
  }
  m_isLocked = false;
  pthread_mutex_unlock(&m_dbLock);
}

WindowDB::BidiTextKeyIterator WindowDB::bidiTextKeyBegin()
{
  return BidiTextKeyIterator(m_bidiTextKeys.begin());
}


WindowDB::BidiTextKeyIterator WindowDB::bidiTextKeyEnd()
{
  return BidiTextKeyIterator(m_bidiTextKeys.end());
}
   
  
WindowDB::DisplayMapIterator 
WindowDB::dpyMapBegin(BidiTextKeyIterator& current)
{
  DisplayMap& dpyMap = (*current.m_it).second.dpyMap;
  return dpyMap.begin();
}


WindowDB::DisplayMapIterator WindowDB::dpyMapEnd(BidiTextKeyIterator& current)
{
  DisplayMap& dpyMap = (*current.m_it).second.dpyMap;
  return dpyMap.end();
}


WindowDB::WindowSetIterator WindowDB::wndsBegin(DisplayMapIterator& current)
{
  WindowSet& wnds = (*current).second;
  return wnds.begin();
}

WindowDB::WindowSetIterator WindowDB::wndsEnd(DisplayMapIterator& current)
{
  WindowSet& wnds = (*current).second;
  return wnds.end();  
}
  


