#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <errno.h>

#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "refreshd_cnt.h"
#include "messages.h"
#include "refreshd_params.h"


#define SIZEOF(arr)   (sizeof(arr)/sizeof(arr[0]))



struct errMsgData {
  RD_CNT_STATUS status;
  const char* message;
};

static struct errMsgData errMsgs[] = {
  {RD_CNT_SUCCESS, 
   "Success"},
  {RD_CNT_UNINITIALIZED_ERROR,
   "Library refreshd_cnt uninitialized"},
  {RD_CNT_NO_DAEMON_ERROR,
   "Refresh daemon does not respond."},
  {RD_CNT_INVALID_BIDITEXT_KEY_ERROR,
   "The biditext key parameter is invalid"},
  {RD_CNT_INVALID_DISPLAY_ERROR,
   "The display parameter is invalid"},
  {RD_CNT_SOCKET_ERROR,
   "Socket creation/binding/connection error"},
  {RD_CNT_SEND_ERROR,
   "Send error"},
  {RD_CNT_RECV_ERROR,
   "Receive error"},
  {RD_CNT_ADDRESS_TOO_LONG_ERROR,
   "Address too long"},
  {RD_CNT_PROTOCOL_ERROR,
   "refreshd protocol error"},
  {RD_CNT_UNKNOWN_ERROR,
   "Unknown error"}
};

const char* rdCntGetErrorMessage(RD_CNT_STATUS status)
{
  size_t i;
  
  for (i=0; i<SIZEOF(errMsgs); i++) {
    const struct errMsgData* emd = &errMsgs[i];
    
    if (emd->status == status)
      return emd->message;
  }
  
  return "Unknown error";
}


static bool libRefreshdCntInitialized = false;
static struct sockaddr_un destAddr;
static RD_CNT_STATUS sendMessage(Message* msg,
				 bool recvAck);


RD_CNT_STATUS rdCntInitialize(void)
{
  pid_t pid = getpid();
  char* ptr = destAddr.sun_path;
  RD_CNT_STATUS errCode = RD_CNT_UNKNOWN_ERROR;
  Message msg;
  
  srand((unsigned int)time(NULL));
  
  memset(&destAddr,0,sizeof(destAddr));
  
  destAddr.sun_family = AF_UNIX;
#ifdef ABSTRACT_UNIX_SOCKET_ADDRESSES
  *ptr++ = '\0';
#endif
  
  if (sizeof(REFRESHD_SOCKET_NAME) >= sizeof(destAddr.sun_path)-1) {
    errCode = RD_CNT_ADDRESS_TOO_LONG_ERROR;
    goto err_addr_too_long;
  }
  
  strcpy(ptr,REFRESHD_SOCKET_NAME);
  
  msg.type = REFRESHD_PING;
  msg.pingMsg.pid = pid;
  
  errCode = sendMessage(&msg,true);
  
  if (errCode != RD_CNT_SUCCESS) {
    errCode = RD_CNT_NO_DAEMON_ERROR;
    goto err_no_daemon;
  }
  
  if (msg.type !=  REFRESHD_REPLY_PING ||
      pid != msg.replyPingMsg.pid) {
    errCode = RD_CNT_PROTOCOL_ERROR;
    goto err_protocol;
  }
  
  libRefreshdCntInitialized = true; 
  
  return RD_CNT_SUCCESS;
  
 err_protocol:
 err_no_daemon:
 err_addr_too_long:
  return errCode;
}


void rdCntFinalize(void)
{
  libRefreshdCntInitialized = false;
}

static void generateLocalAddress(char address[],size_t addrBuffSize)
{
  struct timeval timeStamp;
  int randNum;
  pid_t myPid;
  
  assert(addrBuffSize >= 0);
  assert(MAX_ADDRESS_SIZE + 2 < addrBuffSize);
  
  memset(address,0,addrBuffSize);
  gettimeofday(&timeStamp,NULL);
  
  myPid = getpid();
  randNum = rand();
  
#ifdef ABSTRACT_UNIX_SOCKET_ADDRESSES
  *address++ = '\0';
#endif
  
  sprintf(address,"%s_%d_%ld_%ld_%d",LOCAL_ADDR_PREFIX,myPid,timeStamp.tv_sec,
	  timeStamp.tv_usec,randNum);
}

/* Returns a socket that is bound to a temp address
   and connected to the destination address */

static int getSocket(void)
{
  int sd;
  struct sockaddr_un myAddr;
  int numRetries;
  int rc;
  
  sd = socket(PF_UNIX,SOCK_DGRAM,0);
  
  if (sd < 0) {
    goto err_socket;
  }
  
  myAddr.sun_family = AF_UNIX;
  
  numRetries = 0;
 bind_again:
  generateLocalAddress(myAddr.sun_path,sizeof(myAddr.sun_path));
  
  rc = bind(sd,(struct sockaddr*)&myAddr,sizeof(myAddr));
  
  if (rc < 0) {
    if (numRetries++ < MAX_BIND_RETRIES)
      goto bind_again;
    goto err_bind;
  }
  
  numRetries = 0;
 connect_again:
  rc = connect(sd,(struct sockaddr*)&destAddr,sizeof(destAddr));
  
  if (rc < 0) {
    if (numRetries++ < MAX_CONNECT_RETRIES)
      goto connect_again;
    goto err_connect;
  }
  
  return sd;
  
 err_connect:
#ifndef ABSTRACT_UNIX_SOCKET_ADDRESSES 
  unlink(myAddr.sun_path);
#endif
 err_bind:
  close(sd);
 err_socket:
  return -1;
}

static void destroySocket(int sd)
{
#ifndef ABSTRACT_UNIX_SOCKET_ADDRESSES
  struct sockaddr_un myAddr;
  socklen_t size = sizeof(myAddr);
  
  getsockname(sd,(struct sockaddr*)&myAddr,&size);
  
  unlink(myAddr.sun_path);
#endif
  
  close(sd);
}


static RD_CNT_STATUS sendMessageWithoutAck(int sd,const Message* msg)
{
  int rc;
  int numRetries = 0;
  RD_CNT_STATUS errCode = RD_CNT_UNKNOWN_ERROR;
  
 again:
  rc = send(sd,msg,sizeof(*msg),0);
  
  if (rc < 0) {
    if ( numRetries++ <  MAX_SEND_RETRIES)
      goto again;
    errCode = RD_CNT_SEND_ERROR;
    goto err_send;
  }

  return RD_CNT_SUCCESS;
  
 err_send:
  return errCode;
}


static RD_CNT_STATUS sendMessage(Message* msg,
				 bool recvAck)
{
  int rc;
  int numRetries = 0;
  int sd = -1;
  RD_CNT_STATUS errCode = RD_CNT_UNKNOWN_ERROR;

  sd = getSocket();
  
  if (sd < 0) {
    errCode = RD_CNT_SOCKET_ERROR;
    goto err_socket;
  }
  
  errCode = sendMessageWithoutAck(sd,msg);
  
  if (errCode != RD_CNT_SUCCESS)
    goto err_send;
  
  if (recvAck) {
    fd_set fdset;
    struct timeval timeOut;
    
  select_again:
    FD_ZERO(&fdset);
    FD_SET(sd,&fdset);
    
    timeOut.tv_sec = RECV_TIMEOUT;
    timeOut.tv_usec = 0;
    
    rc = select(sd+1,&fdset,NULL,NULL,&timeOut);
    
    if (rc < 0 ) {
      if (errno == EINTR)
	goto select_again;
      errCode = RD_CNT_RECV_ERROR;
      goto err_select;
    }
    
    if (rc == 0) {
      errCode = RD_CNT_RECV_ERROR;
      goto err_select_timeout;
    }
    
  recv_again:
    rc = recv(sd,msg,sizeof(*msg),0);
    if (rc < 0) {
      if (numRetries++ < MAX_RECV_RETRIES)
	goto recv_again;
      errCode = RD_CNT_RECV_ERROR;
      goto err_recv;
    }
  }
  
  destroySocket(sd);
  
  return RD_CNT_SUCCESS;
  
  
 err_recv:
 err_select_timeout: 
 err_select:
 err_send:
  destroySocket(sd);
 err_socket:
  return errCode;
}

RD_CNT_STATUS rdCntPingDaemon(void)
{
  RD_CNT_STATUS errCode = RD_CNT_UNKNOWN_ERROR;
  Message msg;
  PingMessageData* pmd = &msg.pingMsg;
  pid_t pid = getpid();

  if (!libRefreshdCntInitialized) {
    errCode = RD_CNT_UNINITIALIZED_ERROR;
    goto err_uninitialized;
  }  

  msg.type = REFRESHD_PING;
  pmd->pid = pid;
  errCode = sendMessage(&msg,true);
  if (errCode != RD_CNT_SUCCESS)
    goto err_send_message;

  if (msg.type != REFRESHD_REPLY_PING ||
      msg.replyPingMsg.pid != pid) {
    errCode = RD_CNT_PROTOCOL_ERROR;
    goto err_protocol;
  }
    
  return RD_CNT_SUCCESS;
  
 err_protocol:
 err_send_message:
 err_uninitialized:
  return errCode;
}
  


RD_CNT_STATUS rdCntRefreshNow(void)
{
  RD_CNT_STATUS errCode = RD_CNT_UNKNOWN_ERROR;
  Message msg;
  
  if (!libRefreshdCntInitialized) {
    errCode = RD_CNT_UNINITIALIZED_ERROR;
    goto err_uninitialized;
  }  

  msg.type = REFRESHD_REFRESH_NOW;
  errCode = sendMessage(&msg,false);

  if (errCode != RD_CNT_SUCCESS)
    goto err_send_message;

  return RD_CNT_SUCCESS;
  
 err_send_message:
 err_uninitialized:
  return errCode; 
}

RD_CNT_STATUS rdCntAddWindow(Display* dpy,
			     Window w,
			     const char* bidiTextKey)
{
  RD_CNT_STATUS errCode = RD_CNT_UNKNOWN_ERROR;
  Message msg;
  AddWindowMessageData* awmd = &msg.addWindowMsg;
  pid_t pid = getpid();
  
  if (!libRefreshdCntInitialized) {
    errCode = RD_CNT_UNINITIALIZED_ERROR;
    goto err_uninitialized;
  }
  
  if (bidiTextKey == NULL) {
    errCode = RD_CNT_INVALID_BIDITEXT_KEY_ERROR;
    goto err_invalid_biditext_key;
  }
  
  
  if (dpy == NULL) {
    errCode = RD_CNT_INVALID_DISPLAY_ERROR;
    goto err_invalid_display;
  }
  
  
  msg.type = REFRESHD_ADD_WINDOW;
  
  awmd->bidiTextKeyAtom = XInternAtom(dpy,bidiTextKey,False);
  
  strncpy(awmd->dispName,DisplayString(dpy),MAX_DISP_NAME-1);
  awmd->dispName[MAX_DISP_NAME-1] = '\0';
  
  awmd->pid = pid;
  awmd->w = w;
  
  errCode = sendMessage(&msg,true);
  if (errCode != RD_CNT_SUCCESS)
    goto err_send_recv;
  
  if (msg.type != REFRESHD_REPLY_ADD_WINDOW ||
      msg.replyAddWindowMsg.pid != pid) {
    errCode = RD_CNT_PROTOCOL_ERROR;
    goto err_protocol;
  }
  
  return RD_CNT_SUCCESS;
  
 err_protocol:
 err_send_recv:
 err_invalid_display:
 err_invalid_biditext_key:
 err_uninitialized:
  return errCode;
}


